#include "http_configuration.h"
#include "utils.h"
#include "bgi_lib.h"
#include "builtin.h"
#include <stdio.h>

int http_configuration::load_configuration(char* fstab, char* mime_tab,
					   char* rt, char* log_tab,int pt) {
  port = pt;
  root = strdup(rt);
  FILE* in = fopen(fstab,"r");
  if(in == 0) {
    fprintf(stderr,"can't open configuration file %s;\n ==> ",fstab);
    perror("");
    fclose(in);
    return(-1);
  }
  
  char *line;
  int line_no=0;

  while(line = getline(in,&line_no)) {
    char** items = split_any(line,"\t ",1);
    if(!items) {
      printf("%s:%d: syntax error\n",fstab,line_no);
      delete line;
      fclose(in);
      return(-1);
    }
    char** tmp = items;
    int n=0;
    while(*tmp) {
      ++n;
      ++tmp;
    }
    if(n <3 || n >4) {
      printf("%s:%d: argument missing\n",fstab,line_no);
      delete line;
      delete items;
      fclose(in);
      return(-1);
    }
    if(!strcmp(items[2],"builtin")) {
      int n;
      for(n=0; n <num_builtins;n++) {
	if(!strcmp(builtins[n].name,items[1])) {
	  if(dispatcher->add_handler(items[0],builtins[n].init,
				   builtins[n].umount, builtins[n].handler,
				   items[3]) <0) {
	    printf("%s:%d: couldn't mount builtin \"%s\" \n",
		   fstab, line_no, items[1]);
	    delete line;
	    delete items;
	    fclose(in);
	    return(-1);
	  }
	  break;
	}
      }
      if(n == num_builtins) {
	printf("%s:%d: unknown builtin \"%s\"\n",
	       fstab,line_no,items[1]);
	delete line;
	delete items;
	fclose(in);
	return(-1);
      }
    } else {
      if(dispatcher->add_handler(items[0],items[1],items[2],items[3]) <0) {
	printf("%s:%d: can't mount %s on %s\n",fstab,line_no,
	       items[1],items[0]);
	delete line;
	delete items;
	fclose(in);
	return(-1);
      }
    }
    delete line;
    delete items;
  }
  fclose(in);

  in = fopen(mime_tab,"r");
  if(in == 0) {
    fprintf(stderr,"can't open mime table %s -",mime_tab);
    perror("");
    return(-1);
  }
  line_no =0;
  while(line = getline(in,&line_no)) {
    char** items = split_any(line,"\t ",1);
    if(items && *items) {
      char** tmp = items;
      char* value = *tmp++;
      while(*tmp) {
	(*ext_to_mime_table)[strdup(*tmp)] = strdup(value);
	tmp++;
      }
    } else {
      fprintf(stderr,"%s:%d:  error in extensions table\n",
	      mime_tab,line_no);
    }
    delete items;
    delete line;
  }
  fclose(in);

  in = fopen(log_tab,"r");
  if(in == 0 ) {
    fprintf(stderr,"can't open log table %s\n",log_tab);
    return -1;
  }
  line_no =0;
  while(line = getline(in,&line_no)) {
    char** items = split_any(line,"\t ",1);
    if(items && *items && *(items+1) && *(items+2)) {
      char** tmp = items;
      char* prefix = *tmp++;
      char* library = *tmp++;
      char* args = *tmp;
      if(message_logger->push(prefix,library,args) <0) {
	fprintf(stderr,"%s:%d:  couldn't push log module %s\n",
	      log_tab,line_no,prefix);
      }
    } else {
      fprintf(stderr,"%s:%d:  error in extensions table\n",
	      log_tab,line_no);
    }
    delete items;
    delete line;
  }
  fclose(in);
  return 0;
    
}
