#include "bgi_lib.h"
#include <stdio.h>
#include <utils.h>
#include <sys/time.h>
#include <unistd.h>

extern "C" { int gettimeofday(struct timeval* tv); }

extern "C" {
  void* sample_init(char* path, char* args);
  int sample_umount(char* path, void* cookie);
  int sample_handler(void* cookie, char* method, char* url, char* version,
		   sock_buf* buf);
}


struct sample_cookie {
  void* lock;
  int cnt;
  sample_cookie() {
    lock = create_bgi_mutex();
    cnt = 0;
  }
};

void* sample_init(char* path, char* args) {
  if(debug) {
    printf("mounting sample handler on %s with args %s\n",
	   path,args);
  }
  return new sample_cookie;
}

int sample_umount(char* path, void* cookie) {
  sample_cookie* fc = (sample_cookie*)cookie;
  delete fc;
  return(0);
}

int sample_handler(void* cookie, char* method, char* url , char* version,
		 sock_buf* buf) {
  sample_cookie* fc = (sample_cookie*)cookie;

  lock_bgi_mutex(fc->lock);
  int my_cnt = fc->cnt++;
  unlock_bgi_mutex(fc->lock);
  if(debug) {
    printf("sample handler called to  %s on %s, using version %s\n",
	   method,url,version ? version : "0.9");
  }

  int len;
  struct sockaddr addr; 
  len = sizeof(struct sockaddr);
  struct timeval date;
  gettimeofday(&date);

  if ((getpeername(buf->sock, &addr, &len)) >= 0) {
    log_message(method,url,0,0,200,date.tv_sec,&addr,len);
  }
  
  mime_headers(buf->sock,0,"text/html",0,0);
  char* cp = buf->buffer;
  cp += sprintf(cp,"<H1>How many have I served?</H1>\n");
  cp += sprintf(cp,"The answer is %d\n",my_cnt);
  cp += sprintf(cp,"<p><a href=\"/\"> back to the start</a>\n");
  write(buf->sock,buf->buffer,cp-buf->buffer);
  return 0;
}
  
