#ifndef BGI_H
#define BGI_H

#include "trie.h"
#include "bgi_lib.h"
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>

class bgi_handler {
public:
  init_func init;
  umount_func umount;
  handler_func handler;
  char* path;
  void* cookie;
  bgi_handler(char* p, init_func i, umount_func u, handler_func h) {
    init = i; umount = u; handler = h; path =strdup(p);
  }
  ~bgi_handler() {
    if(cookie) {
      unmount();
    }
    delete path;
  }
  void* initialise(char* args) {
     cookie = init(path,args);
     return cookie;
  }
  int unmount() {
    umount(path,cookie);
    cookie = 0;
  }
  int handle(char* method, char* url, char* version, sock_buf* buf) {
    return handler(cookie,method,url,version,buf);
  }
    
};


class url_dispatcher {
  trie_node<bgi_handler>* name_space;
 public:
  url_dispatcher() {
    name_space = new trie_node<bgi_handler>("",0);
  }
  ~url_dispatcher() {
    delete name_space;
  }
  int add_handler(char* path, init_func i, umount_func u, handler_func h,
		  char* args) {
    bgi_handler* bar = new bgi_handler(path,i,u,h);
    if(bar->initialise(args) == 0) {
      delete bar;
      return -1;
    }
    if(name_space->add(path,bar)) {
      delete bar;
      return -1;
    }
    return 0;
  }
  int add_handler(char* path, char* prefix,char* library,char* args) {
    void* dl_handle = dlopen(library,RTLD_LAZY);
    
    if(dl_handle == 0) {
      printf("open: %s\n",dlerror());
      return -1;
    }
    char buf[1024];
    sprintf(buf,"%s_init", prefix);
    init_func init = (init_func)dlsym(dl_handle, buf);
    if(init == 0) {
      printf("init: %s\n",dlerror());
      dlclose(dl_handle);
      return -1;
    }
    
    sprintf(buf,"%s_umount",prefix);
    umount_func umount = (umount_func)dlsym(dl_handle,buf);
    if(umount == 0) {
      printf("umount: %s\n",dlerror());
      dlclose(dl_handle);
      return -1;
    }

    sprintf(buf,"%s_handler",prefix);
    handler_func handler = (handler_func)dlsym(dl_handle,buf);
    if(handler == 0)       {
      printf("handler: %s\n",dlerror());
      dlclose(dl_handle);
      return -1;
    }
    return add_handler(path,init,umount,handler,args);
  }
    
  int handle(char* method, char* url, char* version, sock_buf* buf) {
    bgi_handler* tmp = name_space->lookup(url);
    if(tmp == 0) {
      return 500;
    } else {
      return tmp->handle(method,url,version,buf);
    }
  }
  
};

#define OP_GET 1
#define OP_PUT 2
#define OP_POST 3
#define OP_HEAD 4

#endif





