#ifndef DLIST_H
#define DLIST_H

#ifdef HAVE_CONFIG_H
#include "mt_config.h"
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_EXCEPTIONS
class dlist_empty {};
class dlist_notfound {};
#endif

template <class C>
class dlist_node  {
public:
  dlist_node*  prev;
  dlist_node*  next;
  C* data;
  char* key;
  dlist_node(char* k, C* d, dlist_node* p=0, dlist_node* n=0) {
    prev = p ; next = n; key = k; data = d;
    if(prev)   prev->next = this;
    if(next)   next->prev = this;
  }
  void unlink() {
    prev->next = next;
    next->prev = prev;
  }
};


template <class C>
class dlist  {
public:
  dlist_node<C>* the_list;
  dlist() {

    the_list = new dlist_node<C>(0,0);
    the_list->next = the_list;
    the_list->prev = the_list;
  };
  ~dlist() {
     the_list->prev->next = 0;
    while(the_list) {
      dlist_node<C>* tmp = the_list->next;
      delete the_list->key;
      delete the_list->data;
      delete the_list;
      the_list = tmp;
    }
  }
  // is the list empty
  int isempty() {
    return the_list == the_list->next;
  }
  // add to start of list;
  void prepend(char* key, C* object) {
    new dlist_node<C>(key,object,the_list,the_list->next);
  }
  // add to end of list
  void append(char* key, C* object) {
    new dlist_node<C>(key,object,the_list->prev,the_list);
  }
  // add to ordered list
  int add(char* key, C* object) {
    dlist_node<C>* cursor = the_list->next;
    while(cursor != the_list) {
      int cmp = strcmp(key,cursor->key);
      if(cmp == 0) {
	delete cursor->data;
	cursor->data = object;
	goto done;
      } else if (cmp < 0) {
	break;
      }
      cursor = cursor->next;
    }
    dlist_node<C>* new_node;
    new_node = new dlist_node<C>(key,object,cursor->prev,cursor);
    new_node->prev->next = new_node;
    new_node->next->prev = new_node;
  done:
    return 0;
  }
  // search for key in ordered list
  C* find(char* key,C* not_found=0) {
    dlist_node<C>* cursor = the_list->next;
    while(cursor != the_list) {
      int cmp = strcmp(key,cursor->key);
      if(cmp == 0) {
	return cursor->data;
      } else {
	if(cmp < 0) {
	  break;
	}
      }
      cursor = cursor->next;
    }
    if(not_found) {
      dlist_node<C>* new_node= 
	new dlist_node<C>(key,not_found,cursor->prev,cursor);
      new_node->prev->next = new_node;
      new_node->next->prev = new_node;
    }
    return not_found;
  }
  // remove key in ordered list
  void remove(char* key) {
    dlist_node<C>* cursor = the_list->next;
    while(cursor != the_list) {
      if(!strcmp(key,cursor->key)) {
	delete cursor->key;
	delete cursor->data;
	cursor->unlink();
	delete cursor;
	return;
      }
    }
  }
    
  // pop the top element from the list
  C* pop() {
    if(isempty()) {
      return 0;
    } else {
      dlist_node<C>* top = the_list->next;
      C* value = top->data;
      top->unlink();
      delete top;
      return value;
    }
  }
  void map(void (*f)(char* k,C* i)) {
    dlist_node<C>* cursor = the_list->next;
    while(cursor != the_list) {
      f(cursor->key,cursor->data);
    }
  }
};

#endif
