// $Id: hash_table.h,v 1.1 1994/05/24 20:47:37 ses Exp $
#ifndef HASH_TABLE_H
#define HASH_TABLE_H

#include "hash.h"
#include <string.h>
#include <memory.h>
#include <malloc.h>
//#include <stream.h>

template<class C>
struct hash_list_node {
  char* key;
  C* item;
  hash_list_node* _cdr;

  hash_list_node(char* k, C* i,hash_list_node* next) {
      key = k ; 
      item = i;
      _cdr = next;
      return;
    }
  ~hash_list_node() {
    if(key) delete(key);
    if(item) delete item;
    if(_cdr) delete _cdr;
  }
  hash_list_node* find(char* k) {
    int cmp = strcmp(key,k);
    if(cmp == 0) {
      return this;
    } else if(cmp > 0) {
      return 0;
    } else if(_cdr == 0) {
      return 0;
    } else {
      return _cdr->find(k);
    }
  }
      
  hash_list_node* insert(char* k, C* i) {
    int cmp = strcmp(k,key);
    if(cmp == 0) {
      item = i;
      return this;
    } else if(cmp <0 ) {
      return new hash_list_node(k,i,this);
    } else if(_cdr == 0) {
      _cdr = new hash_list_node(k,i,0);
      return this;
    } else {
      _cdr = _cdr->insert(k,i);
      return this;
    }
  }
  void map(void (*f)(char* k, C* i)) {
    f(key,item);


    if(_cdr) {
      _cdr->map(f);
    }
  }
  void print() {
  }
};


template<class C> 
class hash_list {
  hash_list_node<C>* body;
 public:
  hash_list() {
    body = 0;
  }
  ~hash_list() {
    if(body) {
      delete body;
    }
  }
  C** find(char* k) {
    C** foo=0;
    if(body) {
      hash_list_node<C>* tmp= body->find(k);
      if(tmp) {
	foo= &tmp->item;
      }
    }
    return foo;
  }
  void insert(char* k, C* i) {
    if(body) {
      body = body->insert(k,i);
    } else {
      body = new hash_list_node<C>(k,i,0);
    }
  }
  void map(void (*f)(char* k, C* i)) {
    if(body) {
      body->map(f);
    }
  }
  void print() {
    if(body) 
      body->print();
  }
  
};

template<class C> 
class hash_table {  
  hash_list<C>* buckets;
  int num_buckets;
  
  public:
    hash_table(int nb) {
      num_buckets = nb;
      buckets = new hash_list<C>[num_buckets];
    }
    ~hash_table() {
      delete buckets;
    }
  void map(void (*f)(char* k, C* i)) {
    int i;
    for(i=0;i<num_buckets;i++) {
      buckets[i].map(f);
    }
  }
  C** _lookup(char* key, int add=0) {
    unsigned long code = hash(key)%num_buckets;

   C** found = buckets[code].find(key);
    if(found) {
      return found;
    } else {
      if(add == 0) {
	return 0;
      }
      buckets[code].insert(strdup(key),0);
      found = buckets[code].find(key);
      return found;
    }
  }
  C*& lookup(char* key) { return *_lookup(key,1);}
  void print() {
    int i;
    for(i=0;i<num_buckets;i++) {
      buckets[i].print();
    }
  }
  C*& operator[](char* k) { return lookup(k);}
  };
#endif
