#ifndef TRIE_H
#define TRIE_H
#ifdef HAVE_CONFIG_H
#include "mt_config.h"
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

template <class C>
class trie_node {
 public:

  char* key;
  C* data;
  trie_node<C>** table;

  void ensure_table() {
    if(table) 
      return;
    else {
      table = new trie_node*[256];
      int i;
      for(i=0;i<256;i++)
	table[i] = 0;
    }
  }

  trie_node(char* k, C* d = 0,trie_node** table_in = 0) {
    key = strdup(k);
    data = d;
    table = 0;
  }

  ~trie_node() {
    if(key)  delete key;
    if(data) delete data;
    if(table) {
      int i;
      for(i=0;i<256;i++) if(table[i])   delete table[i];
      delete table;
    }
  }
  
  int add(char* key_in, C* value) {
    char* in_ptr = key_in;
    char* old_ptr = key;
    
    // compare the new key with the key in this node
    while(*in_ptr && *old_ptr && *in_ptr == *old_ptr) {
      in_ptr++;
      old_ptr++;
    }
    
    // check to see if the keys match exactly
    // if they do, just add the new data value
    if(*old_ptr == 0 && *in_ptr == 0) {
      if(data) {
	delete data;
      }
      data = value;
      return 0;
    }
    
    // check to see if the new_key is a prefix of the old key
    // if they do, we need to split this node, adding a table point to
    // the rest of key
    if(*in_ptr == 0) {
      ensure_table();
      table[*old_ptr] = new trie_node<C>(strdup(old_ptr+1),data,table);
      delete key;
      key = strdup(key_in);
      data = value;
      return 0;
    }

    // if we ran out of trie, we should check the table, and branch appropriate
    if(*old_ptr == 0) {
      ensure_table();
      trie_node* tmp = table[*in_ptr];
      if(tmp) {
	return tmp->add(in_ptr+1, value);
      } else {
	table[*in_ptr] = new trie_node<C>(strdup(in_ptr+1),value);
	return 0;
      }
    }

    // if we get here, we had a mismatch, so we need to split the trie
    // at this point and add two new nodes

    ensure_table();
    table[*in_ptr] = new trie_node<C>(strdup(in_ptr+1),value);
    table[*old_ptr] = new trie_node<C>(strdup(old_ptr+1),data,table);
    delete key;
    data  = 0;
    return 0;
  }
  C* lookup(char* key_in) {
    trie_node<C>* tmp = this;
    C* last_data=0;
    char* in_ptr = key_in;

    while(tmp) {
      char* old_ptr = tmp->key;
      while(*old_ptr && *in_ptr && *old_ptr == *in_ptr) {
	old_ptr++;
	in_ptr++;
      }
      // check to see if there was a mismatch 
      if(*old_ptr) {
	return last_data;
      }

      if(*old_ptr == 0 && tmp->data) {
	last_data = tmp->data;
      }

      if(*in_ptr == 0) {
	  return last_data;
      }
      if(tmp->table) {
	tmp = tmp->table[*in_ptr++];
      } else {
	tmp = 0;
      }
    }
    return last_data;
  }
};
  
#endif
