#include "libmt.h"
#include "list.h"
#include <errno.h>

void* thread_support_proc(void* arg) {
  thread_class* real_class = (thread_class*)arg;
  void* result;
  TRY {
    real_class->before();
    result = real_class->run();
  }
#ifdef HAVE_EXCEPTIONS
  catch(thread_exit* a) {
    result = a->value;
  }
  catch(...) {
    real_class->panic();
  }
#endif
  real_class->after();
  delete real_class;
  return result;
}
 
thread_t start_thread(thread_class* c) {
  thread_t tid;
  
  if(thr_create(NULL,0,thread_support_proc,c,c->options(),&tid)) {
    return 0;
  } else {
    return tid;
  }
}
  
// methods for shared_condition

int shared_condition::signal() {
  lock();
  int result=cond_signal(&the_condition);
  unlock();
  return result;
}

void thread_class::exit(void* result) {
#ifdef HAVE_EXCEPTIONS
  throw(thread_exit(result));
#else
  after();
  delete this;
  thr_exit(result);
#endif
}

int thread_class::set_priority(int level) {
  return thr_setprio(thr_self(),level);
}
void  daemon_thread::before() {
  if(setup()) {
    exit(0);
  }
}

void daemon_thread::after() {
  teardown();
}

void* daemon_thread::run() {
  void* args;
  while(args = get_connection()) {
    start_thread(new_child(args));
    delete args;
  }
  return 0;
}
