#include <sys/types.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <sysent.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/resource.h>

#define MAKE_THREAD(function,argument) \
             (thr_create(NULL,32*1024,function,(void*)argument,THR_BOUND,NULL))

int zap_file_to_fd(int port,char* filename);
void* handle_connection(void* port);
void* use_real_daemon(int port, char* command);
char* get_next_word(char* cp, char* buf);

extern "C" { int open(char* name,int mode); }

char prefix[1024];
int listen_port;
main(int argc, char** argv) {
  if(argc <3) {
    printf("usage: %s prefix port\n",argv[0]);
    exit(1);
  } else {
    strcpy(prefix,argv[1]);
    listen_port = atoi(argv[2]);
  }

  struct rlimit foo;
  getrlimit(RLIMIT_NOFILE,&foo);
  foo.rlim_cur= foo.rlim_max;
  if(setrlimit(RLIMIT_NOFILE,&foo) <0) {
    perror("setrlimit:");
  }

  
  int sock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
  sockaddr_in sin;
  memset(&sin,0,sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = listen_port;

  if(  bind(sock,(sockaddr*)&sin,sizeof(sin)) <0) {
    perror("bind:");
    exit(1);
  }
  if(listen(sock,64) <0) {
    perror("listen:");
    exit(1);
  }
  thr_setprio(thr_self(),100);
//  thr_setconcurrency(50);
  while(1) {
    int a;
    sockaddr foo;
    memset(&foo,0,sizeof(foo));
    a = sizeof(foo);
  loop:
    int i;
    i = accept(sock,&foo,&a);
    if(i<0) {
      perror("accept:");
      sleep(1);
      goto loop;
    }
#ifdef DEBUG
    printf("got a connection!\n");
    printf("current concurrency level is %d\n",thr_getconcurrency());
#endif
    thr_create(NULL,0,handle_connection,(void*)i,THR_DETACHED,NULL);
  }
}



void* handle_connection(void* arg) {
  int port = (int) arg;
  char buf[8192];
  
  char command[1000];
  char argument[1000];
  char version[10];

  int n=read(port,buf,8192);
  if(n <=0) {
    fprintf(stderr,"line too short\n");
    return 0;
  } else  if(n == 8192) {
    fprintf(stderr,"line too long\n");
    return 0;
  }

  char* cp = buf;

  cp = get_next_word(cp,command);
  
  if(cp) {
    cp = get_next_word(cp,argument);
  } else {
    *argument = 0;
  }

  if(cp) {
    get_next_word(cp,version);
  } else {
    *version = 0;
  }
#ifdef DEBUG  
  printf("command = %s, argument = %s, version = <%s>\n",
	  command,argument,version);
#endif
  if(strcmp(command,"GET")) {
    exit(0); // XXX HACK FOR PROFILE
    use_real_daemon(port,buf);
    return 0;
  }
  
  if(!strcmp(argument,"/")) {
    strcpy(argument,"index.html");
  }
  char filename[1024];
  sprintf(filename,"%s/%s",prefix,argument);
  if(zap_file_to_fd(port,filename)) {
    printf("error in zap\n");
  }
  if(close(port) <0) {
    perror("close");
  };
  return 0;
}

char* get_next_word(char* cp, char* buf) {
  while(*cp && !isspace(*cp)) {
    *buf++=*cp++;
  }
  *buf=0;
  if(*cp != ' ') {
    return 0;
  } else {
    return cp+1;
  }
}

int zap_file_to_fd(int port,char* filename) {
  char buf[100];
  struct stat sb;
  long page_size = sysconf(_SC_PAGESIZE);

  if(stat(filename,&sb) <0) {
    printf("stat error for %s\n",filename);
    perror("stat");
    return errno;
  }
  if(sb.st_size == 0 ) {
	return 0;
  }

  thr_setprio(thr_self(),0);
	     
  
  int fd = open(filename,0);
  if(fd <0) {
    perror("open:");
    return errno;
  }
  
  unsigned request_size = (sb.st_size + page_size - 1) & ~page_size;
  char* data = mmap(0,request_size,PROT_READ,MAP_SHARED,fd,0);
  if(data == 0) {
    perror("mmap:");
    return errno;
  }
  close(fd);

//  sprintf(buf,"HHTP/1.0 200 OK\r\n");
//  write(port,buf,strlen(buf));
  
  off_t bytes_left= sb.st_size;
  char* cp = data;
  while(bytes_left) {
    int n = write(port,cp,bytes_left);
    if(n < 0) {
      perror("write:");
      return errno;
    }
    cp +=n;
    bytes_left -= n;
  }
  if(munmap(data,request_size) <0) {
    perror("munmap:");
  }
  return 0;
}

void* use_real_daemon(int port, char* command) {
  printf("punting to real daemon\n");
  close(port);
  return 0;
}
